# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['gaussian', 'circular']

import numpy as np
from ._common import ensure_rng

qr = np.linalg.qr

sym_list = 'A', 'AI', 'AII', 'AIII', 'BDI', 'CII', 'D', 'DIII', 'C', 'CI'


def t(sym):
    """Return the value of time-reversal symmetry squared (1, 0, or -1)"""
    if sym not in sym_list:
        raise ValueError('Non-existent symmetry class.')
    if sym in ('CI', 'AI', 'BDI'):
        return 1
    elif sym in ('CII', 'AII', 'DIII'):
        return -1
    else:
        return 0


def p(sym):
    """Return the value of particle-hole symmetry squared (1, 0, or -1)"""
    if sym not in sym_list:
        raise ValueError('Non-existent symmetry class.')
    if sym in ('D', 'DIII', 'BDI'):
        return 1
    elif sym in ('C', 'CI', 'CII'):
        return -1
    else:
        return 0


def c(sym):
    """Return 1 if the system has chiral symmetry, and 0 otherwise."""
    if (t(sym) and p(sym)) or sym == 'AIII':
        return 1
    else:
        return 0


h_t_matrix = {'AI': [[1]], 'CI': [[0, 1], [1, 0]], 'BDI': [[1, 0], [0, -1]],
              'AII': [[0, 1j], [-1j, 0]],
              'CII': [[0, 0, 1j, 0], [0, 0, 0, 1j],
                      [-1j, 0, 0, 0], [0, -1j, 0, 0]],
              'DIII': [[0, 1j], [-1j, 0]]}
h_p_matrix = {'C': [[0, 1j], [-1j, 0]], 'CI': [[0, 1j], [-1j, 0]],
              'CII': [[0, 0, 1j, 0], [0, 0, 0, -1j],
                      [-1j, 0, 0, 0], [0, 1j, 0, 0]],
              'D': [[1]], 'DIII': [[0, 1], [1, 0]], 'BDI': [[1]]}


def gaussian(n, sym='A', v=1., rng=None):
    """Make a n * n random Gaussian Hamiltonian.

    Parameters
    ----------
    n : int
        Size of the Hamiltonian. It should be even for all the classes
        except A, D, and AI, and in class CII it should be a multiple of 4.
    sym : one of 'A', 'AI', 'AII', 'AIII', 'BDI', 'CII', 'D', 'DIII', 'C', 'CI'
         Altland-Zirnbauer symmetry class of the Hamiltonian.
    v : float
        Variance every degree of freedom of the Hamiltonian. The probaility
        distribution of the Hamiltonian is `P(H) = exp(-Tr(H^2) / 2 v^2)`.
    rng: int or rng (optional)
        Seed or random number generator. If no 'rng' is provided the random
        number generator from numpy will be used.

    Returns
    -------
    h : numpy.ndarray
        A numpy array drawn from a corresponding Gaussian ensemble.

    Notes
    -----
    The representations of symmetry operators are chosen according to
    Phys. Rev. B 85, 165409.

    Matrix indices are grouped first according to orbital number,
    then sigma-index, then tau-index.

    Chiral (sublattice) symmetry C always reads:
        H = -tau_z H tau_z.
    Time reversal symmetry T reads:
        AI: H = H^*.
        BDI: H = tau_z H^* tau_z.
        CI: H = tau_x H^* tau_x.
        AII, CII: H = sigma_y H^* sigma_y.
        DIII: H = tau_y H^* tau_y.
    Particle-hole symmetry reads:
        C, CI: H = -tau_y H^* tau_y.
        CII: H = -tau_z sigma_y H^* tau_z sigma_y.
        D, BDI: H = -H^*.
        DIII: H = -tau_x H^* tau_x.

    This implementation should be sufficiently efficient for large matrices,
    since it avoids any matrix multiplication.
    """
    if sym not in sym_list:
        raise ValueError('Unknown symmetry type.')
    if (c(sym) or t(sym) == -1 or p(sym) == -1):
        if n % 2:
            raise ValueError('Matrix dimension should be even in'
                             ' chosen symmetry class.')
        else:
            tau_z = np.array((n // 2) * [1, -1])
            idx_x = np.arange(n) + tau_z

    if sym == 'CII' and n % 4:
        raise ValueError('Matrix dimension should be a multiple of 4 in'
                         'symmetry class CII.')
    factor = v / np.sqrt(2)

    # define random number generator
    rng = ensure_rng(rng)
    randn = rng.randn

    # Generate a Gaussian matrix of appropriate dtype.
    if sym == 'AI':
        h = randn(n, n)
    elif sym in ('D', 'BDI'):
        h = 1j * randn(n, n)
    else:
        h = randn(n, n) + 1j * randn(n, n)

    # Ensure Hermiticity.
    # TODO: write this as h += h.T.conj() once we rely on numpy >= 1.13.0
    h = h + h.T.conj()

    # Ensure Chiral symmetry.
    if c(sym):
        h -= tau_z.reshape(-1, 1) * h * tau_z
        factor *= 0.5

    # Ensure the necessary anti-unitary symmetry.
    if sym in ('AII', 'DIII'):
        h += tau_z.reshape(-1, 1) * h[idx_x][:, idx_x].conj() * tau_z
        factor /= np.sqrt(2)
    elif sym in ('C', 'CI'):
        h -= tau_z.reshape(-1, 1) * h[idx_x][:, idx_x].conj() * tau_z
        factor /= np.sqrt(2)
    elif sym == 'CII':
        sigma_z = np.array((n // 4) * [1, 1, -1, -1])
        idx_sigma_x = np.arange(n) + 2 * sigma_z
        h += (sigma_z.reshape(-1, 1) * h[idx_sigma_x][:, idx_sigma_x].conj() *
              sigma_z)
        factor /= np.sqrt(2)

    h *= factor

    return h


def circular(n, sym='A', charge=None, rng=None):
    """Make a n * n matrix belonging to a symmetric circular ensemble.

    Parameters
    ----------
    n : int
        Size of the matrix. It should be even for the classes C, CI, CII,
        AII, DIII (either T^2 = -1 or P^2 = -1).
    sym : one of 'A', 'AI', 'AII', 'AIII', 'BDI', 'CII', 'D', 'DIII', 'C', 'CI'
         Altland-Zirnbauer symmetry class of the matrix.
    charge : int or None
        Topological invariant of the matrix. Should be one of 1, -1 in symmetry
        classes D and DIII, should be from 0 to n in classes AIII and BDI,
        and should be from 0 to n / 2 in class CII. If charge is None,
        it is drawn from a binomial distribution with p = 1 / 2.
    rng: int or rng (optional)
        Seed or random number generator. If no 'rng' is passed, the random
        number generator provided by numpy will be used.

    Returns
    -------
    s : numpy.ndarray
        A numpy array drawn from a corresponding circular ensemble.

    Notes
    -----
    The representations of symmetry operators are chosen according to
    Phys. Rev. B 85, 165409, except class D.

    Matrix indices are grouped first according to channel number,
    then sigma-index.

    Chiral (sublattice) symmetry C always reads:
        s = s^+.
    Time reversal symmetry T reads:
        AI, BDI: r = r^T.
        CI: r = -sigma_y r^T sigma_y.
        AII, DIII: r = -r^T.
        CII: r = sigma_y r^T sigma_y.
    Particle-hole symmetry reads:
        CI: r = -sigma_y r^* sigma_y
        C, CII: r = sigma_y r^* sigma_y
        D, BDI: r = r^*.
        DIII: -r = r^*.

    This function uses QR decomposition to probe symmetric compact groups,
    as detailed in arXiv:math-ph/0609050. For a reason as yet unknown, scipy
    implementation of QR decomposition also works for symplectic matrices.
    """
    rng = ensure_rng(rng)
    randn = rng.randn

    if sym not in sym_list:
        raise ValueError('Unknown symmetry type.')
    if (t(sym) == -1 or p(sym) == -1) and n % 2:
        raise ValueError('n must be even in chosen symmetry class.')

    # Prepare a real, complex or symplectic Gaussian matrix
    # Real case.
    if p(sym) == 1:
        h = randn(n, n)
    # Complex case.
    elif p(sym) == 0:
        h = randn(n, n) + 1j * randn(n, n)
    # Symplectic case.
    else:  # p(sym) == -1
        h = randn(n, n) + 1j * randn(n, n)
        tau_z = np.array((n // 2) * [1, -1])
        idx_x = np.arange(n) + tau_z
        h -= tau_z.reshape(-1, 1) * h[idx_x][:, idx_x].conj() * tau_z
        h *= 1j

    # Generate a random matrix with proper electron-hole symmetry.
    # This matrix is a random element of groups O(N), U(N), or USp(N).
    s, h = qr(h)
    if p(sym) == -1 and not np.allclose(np.diag(h, 1)[::2], 0, atol=1e-8):
        raise RuntimeError('QR decomposition symmetry failure.')
    h = np.diag(h).copy()
    h /= np.abs(h)
    s *= h

    # Ensure proper topological invariant in classes D and DIII.
    if sym in ('D', 'DIII') and charge is not None:
        if charge not in (-1, 1):
            raise ValueError('Impossible value of topological invariant.')
        det = np.linalg.det(s)
        if sym == 'DIII':
            det *= (-1) ** (n // 2)
        if (charge > 0) != (det > 0):
            idx = np.arange(n)
            idx[-1] -= 1
            idx[-2] += 1
            s = s[idx]

    # Add the proper time-reversal symmetry:
    if sym in ('AI', 'CI'):
        s = np.dot(s.T, s)
        if sym == 'CI':
            tau_z = np.array((n // 2) * [1, -1])
            idx_x = np.arange(n) + tau_z
            s = 1j * tau_z * s[:, idx_x]
    elif sym == 'AII' or sym == 'DIII':
        tau_z = np.array((n // 2) * [1, -1])
        idx_x = np.arange(n) + tau_z
        s = 1j * np.dot(s.T * tau_z, s[idx_x])

    # Add the chiral symmetry:
    elif sym in ('AIII', 'BDI', 'CII'):
        if sym != 'CII':
            if charge is None:
                diag = 2 * rng.randint(2, size=(n,)) - 1
            elif (0 <= charge <= n) and int(charge) == charge:
                diag = np.array(charge * [-1] + (n - charge) * [1])
            else:
                raise ValueError('Impossible value of topological invariant.')
        else:
            if charge is None:
                diag = 2 * rng.randint(2, size=(n // 2,)) - 1
                diag = np.resize(diag, (2, n // 2)).T.flatten()
            elif (0 <= charge <= n // 2) and int(charge) == charge:
                charge *= 2
                diag = np.array(charge * [-1] + (n - charge) * [1])
            else:
                raise ValueError('Impossible value of topological invariant.')

        s = np.dot(diag * s.T.conj(), s)

    return s



# Copyright 2011-2017 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import keyword
from collections import defaultdict

import numpy as np

import sympy
import sympy.abc
import sympy.physics.quantum
from sympy.core.function import AppliedUndef
from sympy.core.sympify import converter
from sympy.core.core import all_classes as sympy_classes
from sympy.physics.matrices import msigma as _msigma

import warnings

from .._common import reraise_warnings

# TODO: remove when sympy correctly includes MutableDenseMatrix (lol).
sympy_classes = set(sympy_classes) | {sympy.MutableDenseMatrix}

momentum_operators = sympy.symbols('k_x k_y k_z', commutative=False)
position_operators = sympy.symbols('x y z', commutative=False)

pauli = [sympy.eye(2), _msigma(1), _msigma(2), _msigma(3)]

extra_ns = sympy.abc._clash.copy()
extra_ns.update({s.name: s for s in momentum_operators})
extra_ns.update({s.name: s for s in position_operators})
extra_ns.update({'kron': sympy.physics.quantum.TensorProduct,
                 'eye': sympy.eye, 'identity': sympy.eye})
extra_ns.update({'sigma_{}'.format(c): p for c, p in zip('0xyz', pauli)})


# workaroud for https://github.com/sympy/sympy/issues/12060
del extra_ns['I']
del extra_ns['pi']


################  Helpers to handle sympy

def lambdify(expr, locals=None):
    """Return a callable object for computing a continuum Hamiltonian.

    .. warning::
        This function uses ``eval`` (because it calls ``sympy.sympify``), and
        thus should not be used on unsanitized input.

    If necessary, the given expression is sympified using
    `kwant.continuum.sympify`.  It is then converted into a callable object.

    Parameters
    ----------
    expr : str or SymPy expression
        Expression to be converted into a callable object
    locals : dict or ``None`` (default)
        Additional definitions for `~kwant.continuum.sympify`.

    Examples
    --------
    >>> f = lambdify('a + b', locals={'b': 'b + c'})
    >>> f(1, 3, 5)
    9

    >>> ns = {'sigma_plus': [[0, 2], [0, 0]]}
    >>> f = lambdify('k_x**2 * sigma_plus', ns)
    >>> f(0.25)
    array([[ 0.   ,  0.125],
           [ 0.   ,  0.   ]])
    """
    with reraise_warnings(level=4):
        expr = sympify(expr, locals)

    args = [s.name for s in expr.atoms(sympy.Symbol)]
    args += [str(f.func) for f in expr.atoms(AppliedUndef, sympy.Function)]

    return sympy.lambdify(sorted(args), expr)


def sympify(expr, locals=None):
    """Sympify object using special rules for Hamiltonians.

    If `'expr`` is already a type that SymPy understands, it will do nothing
    but return that value. Note that ``locals`` will not be used in this
    situation.

    Otherwise, it is sympified by ``sympy.sympify`` with a modified namespace
    such that

    * the position operators "x", "y" or "z" and momentum operators "k_x",
      "k_y", and "k_z" do not commute,
    * all single-letter identifiers and names of Greek letters (e.g. "pi" or
      "gamma") are treated as symbols,
    * "kron" corresponds to ``sympy.physics.quantum.TensorProduct``, and
      "identity" to ``sympy.eye``,
    * "sigma_0", "sigma_x", "sigma_y", "sigma_z" are the Pauli matrices.

    In addition, Python list literals are interpreted as SymPy matrices.

    .. warning::
        This function uses ``eval`` (because it calls ``sympy.sympify``), and
        thus should not be used on unsanitized input.

    Parameters
    ----------
    expr : str or SymPy expression
        Expression to be converted to a SymPy object.
    locals : dict or ``None`` (default)
        Additional entries for the namespace under which `expr` is sympified.
        The keys must be valid Python variable names.  The values may be
        strings, since they are all are sent through `continuum.sympify`
        themselves before use.  (Note that this is a difference to how
        ``sympy.sympify`` behaves.)

        .. note::
            When a value of `locals` is already a SymPy object, it is used
            as-is, and the caller is responsible to set the commutativity of
            its symbols appropriately.  This possible source of errors is
            demonstrated in the last example below.

    Returns
    -------
    result : SymPy object

    Examples
    --------
        >>> sympify('k_x * A(x) * k_x + V(x)')
        k_x*A(x)*k_x + V(x)     # as valid sympy object

        >>> sympify('k_x**2 + V', locals={'V': 'V_0 + V(x)'})
        k_x**2 + V(x) + V_0

        >>> ns = {'sigma_plus': [[0, 2], [0, 0]]}
        >>> sympify('k_x**2 * sigma_plus', ns)
        Matrix([
        [0, 2*k_x**2],
        [0,        0]])

        >>> sympify('k_x * A(c) * k_x', locals={'c': 'x'})
        k_x*A(x)*k_x
        >>> sympify('k_x * A(c) * k_x', locals={'c': sympy.Symbol('x')})
        A(x)*k_x**2

    """
    stored_value = None

    # if ``expr`` is already a ``sympy`` object we may terminate a code path
    if isinstance(expr, tuple(sympy_classes)):
        if locals:
            warnings.warn('Input expression is already SymPy object: '
                          '"locals" will not be used.',
                          RuntimeWarning,
                          stacklevel=2)

        # We assume that all present functions, like "sin", "cos", will be
        # provided by user during the final evaluation through "params".
        # Therefore we make sure they are defined as AppliedUndef, not built-in
        # sympy types.
        subs = {r: sympy.Function(str(r.func))(*r.args)
                for r in expr.atoms(sympy.Function)}

        return expr.subs(subs)

    # if ``expr`` is not a "sympy" then we proceed with sympifying process
    if locals is None:
        locals = {}

    for k in locals:
        if (not isinstance(k, str)
            or not k.isidentifier() or keyword.iskeyword(k)):
            raise ValueError(
                "Invalid key in 'locals': {}\nKeys must be "
                "identifiers and may not be keywords".format(repr(k)))

    # sympify values of locals before updating it with extra_ns
    locals = {k: sympify(v) for k, v in locals.items()}
    for k, v in extra_ns.items():
        locals.setdefault(k, v)
    try:
        stored_value = converter.pop(list, None)
        converter[list] = lambda x: sympy.Matrix(x)
        hamiltonian = sympy.sympify(expr, locals=locals)
        # if input is for example ``[[k_x * A(x) * k_x]]`` after the first
        # sympify we are getting list of sympy objects, so we call sympify
        # second time to obtain ``sympy`` matrices.
        hamiltonian = sympy.sympify(hamiltonian)
    finally:
        if stored_value is not None:
            converter[list] = stored_value
        else:
            del converter[list]

    return hamiltonian


def make_commutative(expr, *symbols):
    """Make sure that specified symbols are defined as commutative.

    Parameters
    ----------
    expr: sympy.Expr or sympy.Matrix
    symbols: sequace of symbols
        Set of symbols that are requiered to be commutative. It doesn't matter
        of symbol is provided as commutative or not.

    Returns
    -------
    input expression with all specified symbols changed to commutative.
    """
    symbols = [sympy.Symbol(s.name, commutative=False) for s in symbols]
    expr = expr.subs({s: sympy.Symbol(s.name) for s in symbols})
    return expr


def monomials(expr, gens=None):
    """Parse ``expr`` into monomials in the symbols in ``gens``.

    Parameters
    ----------
    expr: sympy.Expr or sympy.Matrix
        Sympy expression to be parsed into monomials.
    gens: sequence of sympy.Symbol objects or strings (optional)
        Generators of monomials. If unset it will default to all
        symbols used in ``expr``.

    Returns
    -------
    dictionary (generator: monomial)

    Examples
    --------
        >>> expr = kwant.continuum.sympify("A * (x**2 + y) + B * x + C")
        >>> monomials(expr, gens=('x', 'y'))
        {1: C, x: B, x**2: A, y: A}
    """
    if gens is None:
        gens = expr.atoms(sympy.Symbol)
    else:
        gens = [sympify(g) for g in gens]

    if not isinstance(expr, sympy.MatrixBase):
        return _expression_monomials(expr, gens)
    else:
        output = defaultdict(lambda: sympy.zeros(*expr.shape))
        for (i, j), e in np.ndenumerate(expr):
            mons = _expression_monomials(e, gens)
            for key, val in mons.items():
                output[key][i, j] += val
        return dict(output)


def _expression_monomials(expr, gens):
    """Parse ``expr`` into monomials in the symbols in ``gens``.

    Parameters
    ----------
    expr: sympy.Expr
        Sympy expr to be parsed.
    gens: sequence of sympy.Symbol
        Generators of monomials.

    Returns
    -------
    dictionary (generator: monomial)
    """
    expr = sympy.expand(expr)
    output = defaultdict(lambda: sympy.Integer(0))
    for summand in expr.as_ordered_terms():
        key = []
        val = []
        for factor in summand.as_ordered_factors():
            symbol, exponent = factor.as_base_exp()
            if symbol in gens:
                key.append(factor)
            else:
                val.append(factor)
        output[sympy.Mul(*key)] += sympy.Mul(*val)

    return dict(output)


################ general help functions

def gcd(*args):
    if len(args) == 1:
        return args[0]

    L = list(args)

    while len(L) > 1:
        a = L[len(L) - 2]
        b = L[len(L) - 1]
        L = L[:len(L) - 2]

        while a:
            a, b = b%a, a

        L.append(b)

    return abs(b)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

cdef extern from "mumps_c_types.h":
    ctypedef int MUMPS_INT

    ctypedef float SMUMPS_REAL
    ctypedef float SMUMPS_COMPLEX

    ctypedef double DMUMPS_REAL
    ctypedef double DMUMPS_COMPLEX

    ctypedef struct mumps_complex:
        pass

    ctypedef struct mumps_double_complex:
        pass

    ctypedef float CMUMPS_REAL
    ctypedef mumps_complex CMUMPS_COMPLEX

    ctypedef double ZMUMPS_REAL
    ctypedef mumps_double_complex ZMUMPS_COMPLEX


cdef extern from "zmumps_c.h":
    ctypedef struct ZMUMPS_STRUC_C:
        MUMPS_INT sym, par, job
        MUMPS_INT comm_fortran
        MUMPS_INT icntl[40]
        ZMUMPS_REAL cntl[15]

        MUMPS_INT n

        MUMPS_INT nz
        MUMPS_INT *irn
        MUMPS_INT *jcn
        ZMUMPS_COMPLEX *a

        MUMPS_INT nrhs, lrhs
        ZMUMPS_COMPLEX *rhs

        MUMPS_INT info[40]
        MUMPS_INT infog[40]
        ZMUMPS_REAL rinfo[40]
        ZMUMPS_REAL rinfog[40]

        MUMPS_INT nz_rhs
        ZMUMPS_COMPLEX *rhs_sparse
        MUMPS_INT *irhs_sparse
        MUMPS_INT *irhs_ptr

        MUMPS_INT size_schur
        MUMPS_INT *listvar_schur
        ZMUMPS_COMPLEX *schur

    cdef void zmumps_c(ZMUMPS_STRUC_C *)



# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import pickle
from io import StringIO
from itertools import zip_longest
import numpy as np
from pytest import raises
from kwant.graph.core import (Graph, NodeDoesNotExistError,
                              EdgeDoesNotExistError, DisabledFeatureError)

def test_empty():
    graph = Graph()
    g = graph.compressed()
    assert not g.twoway
    assert not g.edge_nr_translation
    raises(NodeDoesNotExistError, g.out_neighbors, 0)
    raises(NodeDoesNotExistError, g.has_edge, 0, 0)
    raises(DisabledFeatureError, g.edge_id, 0)
    g = graph.compressed(twoway=True, edge_nr_translation=True)
    assert g.twoway
    assert g.edge_nr_translation
    raises(NodeDoesNotExistError, g.in_neighbors, 0)
    raises(NodeDoesNotExistError, g.out_edge_ids, 0)
    raises(NodeDoesNotExistError, g.in_edge_ids, 0)
    raises(EdgeDoesNotExistError, g.edge_id, 0)

def test_num_nodes():
    graph = Graph()
    assert graph.num_nodes == 0
    graph.num_nodes = 2
    assert graph.num_nodes == 2
    raises(ValueError, graph.__setattr__, 'num_nodes', 1)
    g = graph.compressed()
    assert g.num_nodes == 2

def test_large():
    num_edges = 1000
    graph = Graph()
    for i in range(num_edges):
        graph.add_edge(i, i + 1)
    g = graph.compressed()
    g2 = graph.compressed(twoway=True)
    assert num_edges == g.num_nodes - 1
    for i in range(num_edges):
        assert tuple(g.out_neighbors(i)) == (i + 1,)
        assert tuple(g2.in_neighbors(i + 1)) == (i,)

def check_dot(dot_expect, graph):
    output = StringIO()
    graph.write_dot(output)
    assert output.getvalue() == dot_expect
    output.close()

def test_small():
    g = Graph()
    edges = [(0, 1), (0, 2), (1, 2), (2, 1)]
    for edge in edges:
        g.add_edge(*edge)
    dot_expect = """digraph g {
  0 -> 1;
  0 -> 2;
  1 -> 2;
  2 -> 1;
}
"""
    check_dot(dot_expect, g)
    g = g.compressed(twoway=True)

    for edge_should, edge_is in zip_longest(edges, g):
        assert edge_should == edge_is

    edge_ids = []
    for edge in edges:
        edge_ids.append(g.first_edge_id(*edge))

    assert tuple(g.out_neighbors(0)) == (1, 2)
    assert tuple(g.in_neighbors(0)) == ()
    assert tuple(g.out_edge_ids(0)) == (edge_ids[0], edge_ids[1])
    assert tuple(g.in_edge_ids(0)) == ()

    assert tuple(g.out_neighbors(1)) == (2,)
    assert tuple(g.in_neighbors(1)) == (0, 2)
    assert tuple(g.out_edge_ids(1)) == (edge_ids[2],)
    assert tuple(g.in_edge_ids(1)) == (edge_ids[0], edge_ids[3])

    assert tuple(g.out_neighbors(2)) == (1,)
    assert tuple(g.in_neighbors(2)) == (0, 1)
    assert tuple(g.out_edge_ids(2)) == (edge_ids[3],)
    assert tuple(g.in_edge_ids(2)) == (edge_ids[1], edge_ids[2])

    assert g.has_edge(0, 1)
    g.first_edge_id(0, 1)
    assert not g.has_edge(1, 0)
    raises(IndexError, g.first_edge_id, 1, 0)
    check_dot(dot_expect, g)

def test_negative_node_ids():
    g = Graph()
    raises(ValueError, g.add_edge, 0, -1)

    g = Graph(allow_negative_nodes=True)
    g.add_edge(0, -1)
    g.add_edge(-2, 0)
    raises(ValueError, g.add_edge, -3, -4)
    raises(ValueError, g.compressed)
    g1 = g.compressed(allow_lost_edges=True)
    assert g1.num_px_edges == 1
    assert g1.num_xp_edges == 0
    assert g1.has_edge(0, -1)
    raises(DisabledFeatureError, g1.has_edge, -2, 0)
    assert tuple(g1.out_neighbors(0)) == (-1,)
    g2 = g.compressed(twoway=True)
    assert g2.num_px_edges == 1
    assert g2.num_xp_edges == 1
    assert g2.has_edge(0, -1)
    assert g2.has_edge(-2, 0)
    assert tuple(g2.out_neighbors(0)) == (-1,)
    assert tuple(g2.in_neighbors(0)) == (-2,)

def test_add_edges():
    edges = [(0, 1), (1, 2), (2, 3), (3, 0),
             (0, 4), (1, 4), (2, 4), (3, 4)]

    def fill0(g):
        for edge in edges:
            g.add_edge(*edge)
    def fill1(g):
        g.add_edges(edges)
    def fill2(g):
        g.add_edges(np.array(edges))

    prev_dot = None
    for fill in [fill0, fill1, fill2]:
        g = Graph()
        fill(g)
        g = g.compressed()
        output = StringIO()
        g.write_dot(output)
        dot = output.getvalue()
        if prev_dot is not None:
            assert dot == prev_dot
        prev_dot = dot

def test_edge_ids():
    gr = Graph(allow_negative_nodes=True)
    edges = [(0, -1), (-1, 0), (1, 2), (1, 2), (0, -1), (-1, 0), (-1, 0)]
    for edge_nr, edge in enumerate(edges):
        assert gr.add_edge(*edge) == edge_nr

    g = gr.compressed(twoway=True, edge_nr_translation=True)
    assert g.twoway
    assert g.edge_nr_translation
    assert sorted(g.out_edge_ids(1)) == sorted(g.in_edge_ids(2))
    for edge_id in g.out_edge_ids(1):
        assert g.tail(edge_id) == 1
        assert g.head(edge_id) == 2
    for i, edge_id in enumerate(g.all_edge_ids(0, -1)):
        if i == 0:
            assert edge_id == g.first_edge_id(0, -1)
        assert g.tail(edge_id) == 0
        assert g.head(edge_id) == -1
    assert i == 1
    for i, edge_id in enumerate(g.all_edge_ids(-1, 0)):
        if i == 0:
            assert edge_id == g.first_edge_id(-1, 0)
        assert g.tail(edge_id) == None
        assert g.head(edge_id) == 0
    assert i == 2

    for edge_nr, edge in enumerate(edges):
        if edge[0] < 0: continue
        edge_id = g.edge_id(edge_nr)
        assert edge == (g.tail(edge_id), g.head(edge_id))

    g = gr.compressed(edge_nr_translation=True, allow_lost_edges=True)
    raises(EdgeDoesNotExistError, g.edge_id, 1)


def test_pickle():
    gr = Graph(allow_negative_nodes=True)
    edges = [(0, -1), (-1, 0), (1, 2), (1, 2), (0, -1), (-1, 0), (-1, 0)]
    gr.add_edges(edges)
    g = gr.compressed(twoway=True, edge_nr_translation=True)
    g2 = pickle.loads(pickle.dumps(g))
    s = StringIO('')
    g.write_dot(s)
    s2 = StringIO('')
    g2.write_dot(s2)
    assert s.getvalue() == s2.getvalue()
    assert g.__getstate__() == g2.__getstate__()

    gr = Graph(allow_negative_nodes=False)
    edges = [(0, 1), (1, 2), (1, 2), (0, 2)]
    g = gr.compressed(twoway=False, edge_nr_translation=False)
    g2 = pickle.loads(pickle.dumps(g))
    s = StringIO('')
    g.write_dot(s)
    s2 = StringIO('')
    g2.write_dot(s2)
    assert s.getvalue() == s2.getvalue()
    assert g.__getstate__() == g2.__getstate__()



# -*- coding: utf-8 -*-
# Copyright 2011-2018 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

# This module is imported by plotter.py. It contains all the expensive imports
# that we want to remove from plotter.py

# All matplotlib imports must be isolated in a try, because even without
# matplotlib iterators remain useful.  Further, mpl_toolkits used for 3D
# plotting are also imported separately, to ensure that 2D plotting works even
# if 3D does not.

import warnings
from math import sqrt, pi
import numpy as np

try:
    import matplotlib
    import matplotlib.colors
    import matplotlib.cm
    from matplotlib.figure import Figure
    from matplotlib import collections
    from . import _colormaps
    mpl_available = True
    try:
        from mpl_toolkits import mplot3d
        has3d = True
    except ImportError:
        warnings.warn("3D plotting not available.", RuntimeWarning)
        has3d = False
except ImportError:
    warnings.warn("matplotlib is not available, only iterator-providing "
                  "functions will work.", RuntimeWarning)
    mpl_available = False


# Collections that allow for symbols and linewiths to be given in data space
# (not for general use, only implement what's needed for plotter)
def isarray(var):
    if hasattr(var, '__getitem__') and not isinstance(var, str):
        return True
    else:
        return False


def nparray_if_array(var):
    return np.asarray(var) if isarray(var) else var


if mpl_available:
    class LineCollection(collections.LineCollection):
        def __init__(self, segments, reflen=None, **kwargs):
            super().__init__(segments, **kwargs)
            self.reflen = reflen

        def set_linewidths(self, linewidths):
            self.linewidths_orig = nparray_if_array(linewidths)

        def draw(self, renderer):
            if self.reflen is not None:
                # Note: only works for aspect ratio 1!
                #       72.0 - there is 72 points in an inch
                factor = (self.axes.transData.frozen().to_values()[0] * 72.0 *
                          self.reflen / self.figure.dpi)
            else:
                factor = 1

            super().set_linewidths(self.linewidths_orig *
                                                       factor)
            return super().draw(renderer)


    class PathCollection(collections.PathCollection):
        def __init__(self, paths, sizes=None, reflen=None, **kwargs):
            super().__init__(paths, sizes=sizes, **kwargs)

            self.reflen = reflen
            self.linewidths_orig = nparray_if_array(self.get_linewidths())

            self.transforms = np.array(
                [matplotlib.transforms.Affine2D().scale(x).get_matrix()
                 for x in sizes])

        def get_transforms(self):
            return self.transforms

        def get_transform(self):
            Affine2D = matplotlib.transforms.Affine2D
            if self.reflen is not None:
                # For the paths, use the data transformation but strip the
                # offset (will be added later with offsets)
                args = self.axes.transData.frozen().to_values()[:4] + (0, 0)
                return Affine2D().from_values(*args).scale(self.reflen)
            else:
                return Affine2D().scale(self.figure.dpi / 72.0)

        def draw(self, renderer):
            if self.reflen:
                # Note: only works for aspect ratio 1!
                factor = (self.axes.transData.frozen().to_values()[0] /
                          self.figure.dpi * 72.0 * self.reflen)
                self.set_linewidths(self.linewidths_orig * factor)

            return collections.Collection.draw(self, renderer)


    if has3d:
        # Sorting is optional.
        sort3d = True

        # Compute the projection of a 3D length into 2D data coordinates
        # for this we use 2 3D half-circles that are projected into 2D.
        # (This gives the same length as projecting the full unit sphere.)

        phi = np.linspace(0, pi, 21)
        xyz = np.c_[np.cos(phi), np.sin(phi), 0 * phi].T.reshape(-1, 1, 21)
        # TODO: use np.block once we depend on numpy >= 1.13.
        unit_sphere = np.vstack([
            np.hstack([xyz[0], xyz[2]]),
            np.hstack([xyz[1], xyz[0]]),
            np.hstack([xyz[2], xyz[1]]),
        ])

        def projected_length(ax, length):
            rc = np.array([ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()])
            rc = np.apply_along_axis(np.sum, 1, rc) / 2.

            rs = unit_sphere * length + rc.reshape(-1, 1)

            transform = mplot3d.proj3d.proj_transform
            rp = np.asarray(transform(*(list(rs) + [ax.get_proj()]))[:2])
            rc[:2] = transform(*(list(rc) + [ax.get_proj()]))[:2]

            coords = rp - np.repeat(rc[:2].reshape(-1, 1), len(rs[0]), axis=1)
            return sqrt(np.sum(coords**2, axis=0).max())


        # Auxiliary array for calculating corners of a cube.
        corners = np.zeros((3, 8, 6), np.float_)
        corners[0, [0, 1, 2, 3], 0] = corners[0, [4, 5, 6, 7], 1] = \
        corners[0, [0, 1, 4, 5], 2] = corners[0, [2, 3, 6, 7], 3] = \
        corners[0, [0, 2, 4, 6], 4] = corners[0, [1, 3, 5, 7], 5] = 1.0


        class Line3DCollection(mplot3d.art3d.Line3DCollection):
            def __init__(self, segments, reflen=None, zorder=0, **kwargs):
                super().__init__(segments, **kwargs)
                self.reflen = reflen
                self.zorder3d = zorder

            def set_linewidths(self, linewidths):
                self.linewidths_orig = nparray_if_array(linewidths)

            def do_3d_projection(self, renderer):
                super().do_3d_projection(renderer)
                # The whole 3D ordering is flawed in mplot3d when several
                # collections are added. We just use normal zorder. Note the
                # "-" due to the different logic in the 3d plotting, we still
                # want larger zorder values to be plotted on top of smaller
                # ones.
                return -self.zorder3d

            def draw(self, renderer):
                if self.reflen:
                    proj_len = projected_length(self.axes, self.reflen)
                    args = self.axes.transData.frozen().to_values()
                    # Note: unlike in the 2D case, where we can enforce equal
                    #       aspect ratio, this (currently) does not work with
                    #       3D plots in matplotlib. As an approximation, we
                    #       thus scale with the average of the x- and y-axis
                    #       transformation.
                    factor = proj_len * (args[0] +
                                         args[3]) * 0.5 * 72.0 / self.figure.dpi
                else:
                    factor = 1

                super().set_linewidths(
                                                self.linewidths_orig * factor)
                super().draw(renderer)


        class Path3DCollection(mplot3d.art3d.Patch3DCollection):
            def __init__(self, paths, sizes, reflen=None, zorder=0,
                         offsets=None, **kwargs):
                paths = [matplotlib.patches.PathPatch(path) for path in paths]

                if offsets is not None:
                    kwargs['offsets'] = offsets[:, :2]

                super().__init__(paths, **kwargs)

                if offsets is not None:
                    self.set_3d_properties(zs=offsets[:, 2], zdir="z")

                self.reflen = reflen
                self.zorder3d = zorder

                self.paths_orig = np.array(paths, dtype='object')
                self.linewidths_orig = nparray_if_array(self.get_linewidths())
                self.linewidths_orig2 = self.linewidths_orig
                self.array_orig = nparray_if_array(self.get_array())
                self.facecolors_orig = nparray_if_array(self.get_facecolors())
                self.edgecolors_orig = nparray_if_array(self.get_edgecolors())

                Affine2D = matplotlib.transforms.Affine2D
                self.orig_transforms = np.array(
                    [Affine2D().scale(x).get_matrix() for x in sizes])
                self.transforms = self.orig_transforms

            def set_array(self, array):
                self.array_orig = nparray_if_array(array)
                super().set_array(array)

            def set_color(self, colors):
                self.facecolors_orig = nparray_if_array(colors)
                self.edgecolors_orig = self.facecolors_orig
                super().set_color(colors)

            def set_edgecolors(self, colors):
                colors = matplotlib.colors.colorConverter.to_rgba_array(colors)
                self.edgecolors_orig = nparray_if_array(colors)
                super().set_edgecolors(colors)

            def get_transforms(self):
                # this is exact only for an isometric projection, for the
                # perspective projection used in mplot3d it's an approximation
                return self.transforms

            def get_transform(self):
                Affine2D = matplotlib.transforms.Affine2D
                if self.reflen:
                    proj_len = projected_length(self.axes, self.reflen)

                    # For the paths, use the data transformation but strip the
                    # offset (will be added later with the offsets).
                    args = self.axes.transData.frozen().to_values()[:4] + (0, 0)
                    return Affine2D().from_values(*args).scale(proj_len)
                else:
                    return Affine2D().scale(self.figure.dpi / 72.0)

            def do_3d_projection(self, renderer):
                xs, ys, zs = self._offsets3d

                # numpy complains about zero-length index arrays
                if len(xs) == 0:
                    return -self.zorder3d

                proj = mplot3d.proj3d.proj_transform_clip
                vs = np.array(proj(xs, ys, zs, renderer.M)[:3])

                if sort3d:
                    indx = vs[2].argsort()[::-1]

                    self.set_offsets(vs[:2, indx].T)

                    if len(self.paths_orig) > 1:
                        paths = np.resize(self.paths_orig, (vs.shape[1],))
                        self.set_paths(paths[indx])

                    if len(self.orig_transforms) > 1:
                        self.transforms = self.transforms[indx]

                    lw_orig = self.linewidths_orig
                    if (isinstance(lw_orig, np.ndarray) and len(lw_orig) > 1):
                        self.linewidths_orig2 = np.resize(lw_orig,
                                                           (vs.shape[1],))[indx]

                    # Note: here array, facecolors and edgecolors are
                    #       guaranteed to be 2d numpy arrays or None.  (And
                    #       array is the same length as the coordinates)

                    if self.array_orig is not None:
                        super(Path3DCollection,
                              self).set_array(self.array_orig[indx])

                    if (self.facecolors_orig is not None and
                        self.facecolors_orig.shape[0] > 1):
                        shape = list(self.facecolors_orig.shape)
                        shape[0] = vs.shape[1]
                        super().set_facecolors(
                            np.resize(self.facecolors_orig, shape)[indx])

                    if (self.edgecolors_orig is not None and
                        self.edgecolors_orig.shape[0] > 1):
                        shape = list(self.edgecolors_orig.shape)
                        shape[0] = vs.shape[1]
                        super().set_edgecolors(
                                                np.resize(self.edgecolors_orig,
                                                          shape)[indx])
                else:
                    self.set_offsets(vs[:2].T)

                # the whole 3D ordering is flawed in mplot3d when several
                # collections are added. We just use normal zorder, but correct
                # by the projected z-coord of the "center of gravity",
                # normalized by the projected z-coord of the world coordinates.
                # In doing so, several Path3DCollections are plotted probably
                # in the right order (it's not exact) if they have the same
                # zorder. Still, smaller and larger integer zorders are plotted
                # below or on top.

                bbox = np.asarray(self.axes.get_w_lims())

                proj = mplot3d.proj3d.proj_transform_clip
                cz = proj(*(list(np.dot(corners, bbox)) + [renderer.M]))[2]

                return -self.zorder3d + vs[2].mean() / cz.ptp()

            def draw(self, renderer):
                if self.reflen:
                    proj_len = projected_length(self.axes, self.reflen)
                    args = self.axes.transData.frozen().to_values()
                    factor = proj_len * (args[0] +
                                         args[3]) * 0.5 * 72.0 / self.figure.dpi

                    self.set_linewidths(self.linewidths_orig2 * factor)

                super().draw(renderer)



# Copyright 2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Random-access random numbers

This module provides routines that given some input compute a "random" output
that depends on the input in a (cryptographically) intractable way.

This turns out to be very useful when one needs to map some irregular objects to
random numbers in a deterministic and reproducible way.

Internally, the md5 hash algorithm is used.  The randomness thus generated is
good enough to pass the "dieharder" battery of tests: see the function `test` of
this module.
"""

from math import pi, log, sqrt, cos
from hashlib import md5
from struct import unpack

__all__ = ['uniform', 'gauss', 'test']


TWOPI = 2 * pi
BPF = 53                        # Number of bits in a Python float.
BPF_MASK = 2**53 - 1
RECIP_BPF = 2**-BPF


def str_to_bytes(s):
    """Return bytes if the input is a string, else return the object as-is."""
    try:
        return s.encode('utf8')
    except AttributeError:
        return s

def uniform2(input, salt=''):
    """Return two independent [0,1)-distributed numbers."""
    input = str_to_bytes(input)
    salt = str_to_bytes(salt)
    input = memoryview(input).tobytes() + salt
    a, b = unpack('qq', md5(input).digest())
    a &= BPF_MASK
    b &= BPF_MASK
    return a * RECIP_BPF, b * RECIP_BPF


def uniform(input, salt=''):
    """md5-hash `input` and `salt` and map the result to the [0,1) interval.

    `input` must be some object that supports the buffer protocol (i.e. a string
    or a numpy/tinyarray array).  `salt` must be a string or a bytes object.
    """
    return uniform2(input, salt)[0]


def gauss(input, salt=''):
    """md5-hash `input` and `salt` and return the result as a standard normal
    distributed variable.

    `input` must be some object that supports the buffer protocol (i.e. a string
    or a numpy/tinyarray array).  `salt` must be a string or a bytes object.
    """
    # This uses the Box-Muller transform.  Only one of the two results is
    # computed.
    a, b = uniform2(input, salt)
    return cos(a * TWOPI) * sqrt(-2.0 * log(1.0 - b))


def test(n=20000):  # skip coverage
    """Test the generator with the dieharder suite generating n**2 samples.

    Executing this function may take a very long time.
    """
    import os
    import tempfile
    import subprocess
    from tinyarray import array
    from struct import pack

    f = tempfile.NamedTemporaryFile(delete=False)
    try:
        for x in range(n):
            for y in range(n):
                a = array((x, y))
                i = int(2**32 * uniform(a))
                f.write(pack('I', i))
        f.close()
        subprocess.call(['dieharder', '-a', '-g', '201', '-f', f.name])
    finally:
        os.remove(f.name)



# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

from copy import copy as copy
from types import SimpleNamespace

import pytest
import numpy as np
import scipy.sparse
import scipy.sparse.linalg as sla
from scipy.integrate import simps

import kwant
from ..kpm import _rescale, LocalVectors, RandomVectors
from .._common import ensure_rng


SpectralDensity = kwant.kpm.SpectralDensity

# ### General parameters
dim = 20
p = SimpleNamespace(
    num_moments=200,
    num_vectors=5,
    )
ham = kwant.rmt.gaussian(dim)

# ### Auxiliary functions
TOL = 1e-12
TOL_SP = 1e-6
TOL_WEAK = 1e-3


def assert_allclose(arr1, arr2):
    np.testing.assert_allclose(arr1, arr2, rtol=0., atol=TOL)


def assert_allclose_sp(arr1, arr2):
    np.testing.assert_allclose(arr1, arr2, rtol=0., atol=TOL_SP)

def test_mean_SD():
    kpm_params = dict(num_vectors=1, num_moments=10, rng=0)
    sites = [np.random.randint(dim), np.random.randint(dim)]
    local_spectrum = SpectralDensity(
        ham, vector_factory=LocalVectors(ham, where=sites),
        mean=False, **kpm_params)
    spectrum = SpectralDensity(
        ham, vector_factory=LocalVectors(ham, where=sites),
        mean=True, **kpm_params)
    assert_allclose(local_spectrum.energies, spectrum.energies)
    assert_allclose(local_spectrum.densities.sum(axis=1), spectrum.densities)

def test_conductivity():
    radius = 3
    letters = [('x','x'), ('x','y'), ('y','x'), ('y','y')]
    syst = kwant.Builder()
    lat = kwant.lattice.square(norbs=1)
    syst._lattice = lat
    def circle(x):
        return np.linalg.norm(x) < radius
    syst[lat.shape(circle, (0, 0))] = 0
    syst[lat.neighbors()] = -1
    syst = syst.finalized()
    hamiltonian = syst.hamiltonian_submatrix()
    positions = np.array([s.pos for s in syst.sites])

    # results for num_moments=10, num_vectors=10
    kpm_params = dict(
        params=None, num_vectors=10, num_moments=12,
        energy_resolution=None, vector_factory=None, bounds=None,
        eps=0.05, rng=0, accumulate_vectors=False)

    # test system and no positions
    for alpha, beta in letters:
        cond = kwant.kpm.conductivity(syst, alpha=alpha, beta=beta,
                                        positions=None, **kpm_params)

    # test system or hamiltonian, no positions, but velocity operators
    cond_xx = kwant.kpm.conductivity(syst, alpha='x', beta='x',
                                     positions=None, **kpm_params)
    x_op = scipy.sparse.coo_matrix(hamiltonian, copy=True)
    displacements = positions[x_op.col] - positions[x_op.row]
    x_op.data *= 1j * displacements[:, 0]
    for ham, pos in [(syst, None), (hamiltonian, positions)]:
        # test formatting of alpha and beta
        for alpha, beta in [('x',x_op), (x_op, 'x'), (x_op, x_op)]:
            cond = kwant.kpm.conductivity(ham, alpha=alpha, beta=beta,
                                          positions=pos, **kpm_params)
            assert_allclose(cond(), cond_xx())

    # test increase number of moments
    increase_moments = 90
    kpm_params['accumulate_vectors'] = True
    cond = kwant.kpm.conductivity(syst, alpha='x', beta='x', positions=None,
                                  **kpm_params)
    cond.add_moments(num_moments=increase_moments)

    # assert that it's equivalent to construct the instance directly
    kpm_params['accumulate_vectors'] = False
    kpm_params['num_moments'] = kpm_params['num_moments'] + increase_moments
    cond_100 = kwant.kpm.conductivity(syst, alpha='x', beta='x', positions=None,
                                      **kpm_params)
    assert_allclose(cond_100(), cond())

def test_where():
    radius = 3
    syst = kwant.Builder()
    lat = kwant.lattice.honeycomb(norbs=1)
    syst._lattice = lat
    def circle(x):
        return np.linalg.norm(x) < radius
    syst[lat.shape(circle, (0, 0))] = 0
    syst[lat.neighbors()] = -1
    syst = syst.finalized()

    where = lambda s: np.linalg.norm(s.pos) < 2
    sites = [s for s in syst.sites if where(s)]

    kpm_params = dict(num_vectors=None, num_moments=10, rng=0)
    spectrum2 = SpectralDensity(
        syst, vector_factory=LocalVectors(syst, where=where), **kpm_params)
    spectrum1 = SpectralDensity(
        syst, vector_factory=LocalVectors(syst, where=sites), **kpm_params)
    assert_allclose(spectrum1(), spectrum2())

    # test local vectors
    cond_xy1 = kwant.kpm.conductivity(
        syst, vector_factory=LocalVectors(syst, where=sites),
        alpha='x', beta='y', **kpm_params)
    cond_xy2 = kwant.kpm.conductivity(
        syst, vector_factory=LocalVectors(syst, where=where),
        alpha='x', beta='y', **kpm_params)
    assert_allclose(cond_xy1(), cond_xy2())

    # test random factory
    kpm_params = dict(num_vectors=10, num_moments=10, rng=0)
    cond_xy1 = kwant.kpm.conductivity(
        syst, vector_factory=RandomVectors(syst, sites, kpm_params['rng']),
        alpha='x', beta='y', **kpm_params)
    cond_xy2 = kwant.kpm.conductivity(
        syst, vector_factory=RandomVectors(syst, where, kpm_params['rng']),
        alpha='x', beta='y', **kpm_params)
    assert_allclose(cond_xy1(), cond_xy2())

def test_vector_factory():

    vectors = np.identity(dim)
    def vectors2():
        count = 0
        while count < 6:
            yield 'string'[count]
            count += 1

    vf = lambda n, v: kwant.kpm._VectorFactory(num_vectors=n, vectors=v)
    this_vf = vf(None, vectors)
    assert this_vf.num_vectors == dim
    this_vf = vf(3, vectors)
    assert this_vf.num_vectors == 3
    this_vf.add_vectors(num_vectors=2)
    assert this_vf.num_vectors == 5

    spectrum = lambda n, v : SpectralDensity(ham, vector_factory=v,
                                             num_vectors=n)
    corr = lambda n, v : kwant.kpm.Correlator(ham, vector_factory=v,
                                              num_vectors=n)

    for instance in [spectrum, corr]:
        this_instance = instance(None, vectors)
        assert this_instance.num_vectors == dim
        this_instance = instance(3, vectors)
        assert this_instance.num_vectors == 3
        this_instance.add_vectors(num_vectors=2)
        assert this_instance.num_vectors == 5


def make_spectrum(ham, p, operator=None, vector_factory=None, rng=None, params=None):
    """Create an instance of SpectralDensity class."""
    return SpectralDensity(
        ham,
        operator=operator,
        vector_factory=vector_factory,
        num_moments=p.num_moments,
        num_vectors=p.num_vectors,
        rng=rng,
        params=params
        )


def make_chain(r=dim, t=-1):
    syst = kwant.Builder()
    lat = kwant.lattice.chain(norbs=1)
    for i in range(r):
        syst[lat(i)] = 0
    syst[lat.neighbors()] = t
    return syst.finalized()


def make_chain_with_params(r=10, pot=0, t=-1):
    syst = kwant.Builder()
    lat = kwant.lattice.chain(norbs=1)
    pot = lambda site, pot: pot
    hop = lambda site1, site2, t: t
    for i in range(r):
        syst[lat(i)] = pot
    syst[lat.neighbors()] = hop
    return syst.finalized()


def kpm_derivative(spectrum, e, order=1):
    """Calculate the coeficients for Chebyshev expansion and its derivates."""
    e = np.asarray(e, dtype=complex)
    rescaled_energy = (e - spectrum._b) / spectrum._a
    g_e = np.pi * np.sqrt(1 - rescaled_energy) * np.sqrt(1 + rescaled_energy)

    m = np.arange(spectrum.num_moments)
    kernel = ((spectrum.num_moments - m + 1) *
              np.cos(np.pi * m / (spectrum.num_moments + 1)) +
              np.sin(np.pi * m / (spectrum.num_moments + 1)) /
              np.tan(np.pi / (spectrum.num_moments + 1)))
    kernel = kernel / (spectrum.num_moments + 1)

    moments = np.sum(spectrum._moments_list, axis=0) /\
        (spectrum.num_vectors * spectrum.hamiltonian.shape[0])
    coef_cheb = np.zeros_like(moments)
    coef_cheb[0] = moments[0]
    coef_cheb[1:] = 2 * moments[1:] * kernel[1:]

    i = 1
    while i <= order:
        i += 1
        coef_cheb = np.polynomial.chebyshev.chebder(coef_cheb)

    return np.polynomial.chebyshev.chebval(rescaled_energy, coef_cheb)/g_e


def make_spectrum_and_peaks(ham, precise, threshold=0.005):
    """Calculate the spectrum and the peaks in the array energies."""
    spectrum = make_spectrum(ham, precise)
    derivate_array = kpm_derivative(spectrum, spectrum.energies)
    second_derivate_array = kpm_derivative(spectrum,
                                           spectrum.energies, order=2)
    # where the derivative changes sign
    zero_index = np.where(np.diff(np.sign(derivate_array)))[0]
    scale = threshold * np.max(spectrum.densities)
    # where the spectrum is avobe the threshold
    filter_index = zero_index[
        np.where(spectrum.densities[zero_index] > scale)[0]]
    # where is a local maxmimum
    filter_index2 = filter_index[
        np.where(second_derivate_array[filter_index] < 0)[0]]
    return spectrum, filter_index2


def deviation_from_eigenvalues(dim, precise):
    """Return the maxmimum difference between the peaks in the spectrum and
    the eigenvalues."""
    ham = kwant.rmt.gaussian(dim)
    spectrum, filter_index = make_spectrum_and_peaks(ham, precise)
    try:
        return 1./(2 * spectrum._a) * np.max(np.abs(
            spectrum.energies[filter_index] - np.linalg.eigh(ham)[0]))
    except ValueError:
        """This means that two eigenvalues are nearly degenerate, therefore the
        maximum deviation is the minimum difference between a peak and all the
        eigenvalues."""
        return np.max([1./(2*spectrum._a) *
                       np.min(np.abs(spectrum.energies[filter_index] -
                                     np.linalg.eigh(ham)[0][i]))
                       for i in range(len(ham))])


def find_peaks_in_an_array(array, threshold=0.005):
    """Find the peaks above threshold and return the index inside array."""
    derivate_array = np.gradient(array)
    second_derivate_array = np.gradient(derivate_array)
    # where the derivative changes sign
    zero_index = np.where(np.diff(np.sign(derivate_array)))[0]
    scale = threshold * np.max(array)
    # where the spectrum is avobe the threshold
    filter_index = zero_index[np.where(array[zero_index] > scale)[0]]
    # where is a local maxmimum
    filter_index2 = filter_index[
        np.where(second_derivate_array[filter_index] < 0)[0]]
    return filter_index2


# ### Tests for consistency in top-level API ###


def test_api_ham():
    not_a_hamiltonian = str('not a hamiltonian')
    with pytest.raises(ValueError):
        SpectralDensity(not_a_hamiltonian)


def test_api_operator():
    not_an_operator = str('not an operator')
    with pytest.raises(ValueError):
        SpectralDensity(kwant.rmt.gaussian(dim),
                        operator=not_an_operator)


def test_api_single_eigenvalue_error():
    with pytest.raises(ValueError):
        SpectralDensity(np.identity(dim, dtype=complex))


def test_energy_resolution():
    """Check that energy resolution works and gives the same output as
    setting the equivalent number of moments.
    """
    ham = kwant.rmt.gaussian(dim)
    rng = 1
    energy_resolution = 0.05

    sp1 = SpectralDensity(ham, rng=rng, energy_resolution=energy_resolution)
    # number of moments for this energy resolution
    num_moments = sp1.num_moments
    # use the same number of moments but not passing energy resolution
    sp2 = SpectralDensity(ham, rng=rng, num_moments=num_moments)

    # Check bit for bit equality
    assert np.all(sp1.densities == sp2.densities)


def test_operator_none():
    """Check operator=None gives the same results as operator=np.identity(),
    with the same random vectors.
    """
    ham = kwant.rmt.gaussian(dim)
    identity = np.identity(dim)

    sp1 = SpectralDensity(ham, operator=None, rng=1)
    sp2 = SpectralDensity(ham, operator=identity, rng=1)

    # different algorithms are used so these arrays are equal up to TOL_SP
    assert_allclose_sp(sp1.densities, sp2.densities)


def test_bounds():
    """Check operator=None gives the same results as operator=np.identity(),
    with the same random vectors.
    """
    ham = kwant.rmt.gaussian(dim)
    epsilon = 0.05
    tol = epsilon*0.5
    rng = ensure_rng(1)
    sp1 = SpectralDensity(ham, bounds=None, eps=epsilon, rng=rng)
    # re initialize to obtain the same vector v0
    rng = ensure_rng(1)
    v0 = np.exp(2j * np.pi * rng.random_sample(dim))
    lmax = float(sla.eigsh(
        ham, k=1, which='LA', return_eigenvectors=False, tol=tol, v0=v0))
    lmin = float(sla.eigsh(
        ham, k=1, which='SA', return_eigenvectors=False, tol=tol, v0=v0))
    sp2 = SpectralDensity(ham, bounds=(lmin, lmax), eps=epsilon, rng=1)

    # different algorithms are used so these arrays are equal up to TOL_SP
    assert_allclose_sp(sp1.densities, sp2.densities)


def test_operator_user():
    """Check operator=None gives the same results as operator=np.identity(),
    with the same random vectors.
    """
    ham = kwant.rmt.gaussian(dim)
    identity = np.identity(dim)
    user_op = lambda bra, ket: np.vdot(bra, np.dot(identity, ket))

    sp1 = SpectralDensity(ham, operator=user_op, rng=1)
    sp2 = SpectralDensity(ham, operator=identity, rng=1)

    # different algorithms are used so these arrays are equal up to TOL_SP
    assert_allclose_sp(sp1.densities, sp2.densities)


def test_kwant_syst():
    """Check that when a kwant system is passed, the results are the same as
    for the Hamiltonian (in both sparse and dense formats) of the same kwant
    system.
    """
    syst = make_chain()
    spectrum_syst = make_spectrum(syst, p, rng=1)
    ham_sparse = syst.hamiltonian_submatrix(sparse=True)
    ham_dense = syst.hamiltonian_submatrix(sparse=False)

    spectrum_sparse = make_spectrum(ham_sparse, p, rng=1)
    spectrum_dense = make_spectrum(ham_dense, p, rng=1)

    # same algorithms are used so these arrays are equal up to TOL
    assert_allclose(spectrum_syst.densities, spectrum_sparse.densities)
    assert_allclose(spectrum_syst.densities, spectrum_dense.densities)


def test_kwant_op():
    """Check that the kwant.operator.Density gives the same result as the
    identity operator.
    """
    syst = make_chain()
    kwant_op = kwant.operator.Density(syst)
    spectrum_syst = make_spectrum(syst, p, operator=kwant_op, rng=1)

    ham = syst.hamiltonian_submatrix()
    identity = np.identity(dim)
    spectrum = make_spectrum(ham, p, operator=identity, rng=1)

    assert spectrum_syst.densities.shape[1] == ham.shape[0]
    # same algorithms are used so these arrays are equal up to TOL
    assert_allclose(np.sum(spectrum_syst.densities, axis=1),
                    spectrum.densities)
    # test that the densities are arrays when calling the instance
    e = spectrum_syst.energies
    assert_allclose_sp(spectrum_syst(e), spectrum_syst.densities)


def test_kwant_op_current():
    """Check that the kwant.operator.Density gives the same result as the
    identity operator when the system has ``params``.
    """
    params = {'r':dim, 'pot':1, 't':-2}

    # build the system using parameters default values
    # to be later rewritten
    syst = make_chain_with_params()
    current = kwant.operator.Current(syst)
    # pass parameters to kpm to evaluate hamiltonian and operator
    spectrum_syst = make_spectrum(syst, p, operator=current, rng=1,
                                  params=params)

    # compare with explicit hamiltonian and operator
    ham = syst.hamiltonian_submatrix(params=params)
    def my_current(bra, ket):
        return current(bra, ket, params=params)
    # do not pass parameters to kpm
    spectrum = make_spectrum(ham, p, operator=my_current, rng=1)

    # same algorithms are used so these arrays are equal up to TOL
    assert_allclose(spectrum_syst.densities, spectrum.densities)


def test_kwant_op_integrate():
    """Check that the kwant.operator.Density gives the same result as the
    identity operator.
    """
    syst = make_chain()
    kwant_op = kwant.operator.Density(syst)
    spectrum_syst = make_spectrum(syst, p, operator=kwant_op, rng=1)

    ham = syst.hamiltonian_submatrix()
    identity = np.identity(dim)
    spectrum = make_spectrum(ham, p, operator=identity, rng=1)
    ones = lambda x: np.ones_like(x)

    assert spectrum_syst.densities.shape[1] == ham.shape[0]
    # same algorithms are used so these arrays are equal up to TOL
    assert_allclose(np.sum(spectrum_syst.integrate(distribution_function=ones)),
                    spectrum.integrate())


# ## test for methods to work as expected

# ### increase num_moments


def test_increase_num_moments():
    spectrum = make_spectrum(ham, p, rng=1)
    precise = copy(p)
    precise.num_moments = 2 * p.num_moments

    spectrum_raise = make_spectrum(ham, precise, rng=1)
    spectrum.add_moments(p.num_moments)

    # Check bit for bit equality
    assert np.all(np.array(spectrum_raise._moments_list) ==
                  np.array(spectrum._moments_list))

    # test case when there are an odd number of moments
    # (this code path is treated differently internally)
    odd = copy(p)
    odd.num_moments = p.num_moments + 1
    spectrum_odd = make_spectrum(ham, odd, rng=1)

    assert not np.all(np.asarray(spectrum_odd._moments_list)[:, -1] == 0)

    # test when increasing num_moments by 1 from an even to an odd number
    spectrum_even = make_spectrum(ham, p, rng=1)
    spectrum_even.add_moments(num_moments=1)
    assert np.all(np.array(spectrum_even._moments_list) ==
                  np.array(spectrum_odd._moments_list))


# ### increase num_moments with an operator (different algorithm)

def test_increase_num_moments_op():
    ham = kwant.rmt.gaussian(dim)
    identity = np.identity(dim)

    sp1 = make_spectrum(ham, p, operator=None, rng=1)
    sp2 = make_spectrum(ham, p, operator=identity, rng=1)

    sp1.add_moments(num_moments=sp1.num_moments)
    sp2.add_moments(num_moments=sp2.num_moments)

    # different algorithms are used so these arrays are equal up to TOL_SP
    assert_allclose_sp(sp1.densities, sp2.densities)

# ### increase num_random_vecs


def test_increase_num_vectors():
    precise = copy(p)
    precise.num_vectors = 2 * p.num_vectors

    spectrum_raise = make_spectrum(ham, precise, rng=1)
    spectrum = make_spectrum(ham, p, rng=1)
    spectrum.add_vectors(num_vectors=p.num_vectors)
    # Check bit for bit equality

    assert np.all(spectrum_raise.densities == spectrum.densities)


def test_invalid_input():

    with pytest.raises(TypeError):
        SpectralDensity(ham, num_moments=10, energy_resolution=0.1)

    s = SpectralDensity(ham)
    with pytest.raises(TypeError):
        s.add_moments(num_moments=20, energy_resolution=0.1)

    for v in (-20, 0, 10.5):
        for p in ('num_vectors', 'num_moments'):
            with pytest.raises(ValueError):
                SpectralDensity(ham, **{p: v})
        s = SpectralDensity(ham)
        with pytest.raises(ValueError):
            s.add_moments(num_moments=v)
        with pytest.raises(ValueError):
            s.add_vectors(num_vectors=v)

    for v in (-1, -0.5, 0):
        with pytest.raises(ValueError):
            SpectralDensity(ham, eps=v)
        s = SpectralDensity(ham)
        with pytest.raises(ValueError):
            s.add_moments(energy_resolution=v)

    with pytest.raises(ValueError):
        s.add_moments(energy_resolution=10)


# ### Convergence for higher number of moments ###


def test_check_convergence_decreasing_values():
    difference_list = []
    depth = 2

    for i in range(depth):
        precise = SimpleNamespace(
            num_moments=10*dim + 100*i*dim,
            num_vectors=dim//2,
            )
        results = []
        np.random.seed(1)
        iterations = 3

        for ii in range(iterations):
            results.append(deviation_from_eigenvalues(dim, precise))

        difference_list.append(results)

    assert(np.all(np.diff(np.sum(difference_list, axis=1) / iterations) < 0.))
    assert(np.sum(difference_list, axis=1)[-1] / iterations < TOL_WEAK)

# ### Convergence for a custom vector_factory


def test_convergence_custom_vector_factory():
    rng = ensure_rng(1)

    def random_binary_vectors():
        while True:
            yield np.rint(rng.random_sample(dim)) * 2 - 1

    # extracted from `deviation_from_eigenvalues
    def deviation(ham, spectrum):
        filter_index = find_peaks_in_an_array(spectrum.densities)
        try:
            return 1./(2 * spectrum._a) * np.max(np.abs(
                spectrum.energies[filter_index] - np.linalg.eigh(ham)[0]))
        except ValueError:
            """This means that two eigenvalues are nearly degenerate, therefore
            the maximum deviation is the minimum difference between a peak and
            all the eigenvalues."""

            return np.max([1./(2 * spectrum._a) *
                           np.min(np.abs(spectrum.energies[filter_index] -
                                         np.linalg.eigh(ham)[0][i]))
                           for i in range(len(ham))])

    difference_list = []
    depth = 2

    for i in range(depth):
        precise = SimpleNamespace(
            num_moments=10*dim + 100*i*dim,
            num_vectors=dim//2,
            )
        results = []
        iterations = 3

        for ii in range(iterations):
            ham = kwant.rmt.gaussian(dim)
            spectrum = make_spectrum(ham, precise,
                                     vector_factory=random_binary_vectors())
            results.append(deviation(ham, spectrum))

        difference_list.append(results)

    assert np.all(np.diff(np.sum(difference_list, axis=1) / iterations) < 0.)
    assert np.sum(difference_list, axis=1)[-1]/iterations < TOL_WEAK

# ## Consistency checks for internal functions

# ### check call


def test_call_no_argument():
    ham = kwant.rmt.gaussian(dim)
    spectrum = make_spectrum(ham, p)
    energies, densities = spectrum()

    # different algorithms are used so these arrays are equal up to TOL_SP
    assert_allclose_sp(energies, spectrum.energies)
    assert_allclose_sp(densities, spectrum.densities)


def test_call():
    ham = kwant.rmt.gaussian(dim)
    spectrum = make_spectrum(ham, p)
    densities_array = np.array([spectrum(e) for e in spectrum.energies])

    # different algorithms are used so these arrays are equal up to TOL_SP
    assert_allclose_sp(densities_array, spectrum.densities)

# ### check integrate


def test_integrate():
    ham = kwant.rmt.gaussian(dim)
    spectrum = make_spectrum(ham, p)
    ones = lambda x: np.ones_like(x)
    assert np.abs(
        (spectrum.integrate() - simps(spectrum.densities, x=spectrum.energies))
        / spectrum.integrate()) < TOL_SP
    assert np.abs(spectrum.integrate() - spectrum.integrate(
        distribution_function=ones)) < TOL

# ### check increase_energy_resolution


def test_increase_energy_resolution():
    spectrum, filter_index = make_spectrum_and_peaks(ham, p)

    old_sampling_points = 2 * spectrum.num_moments
    tol = np.max(np.abs(np.diff(spectrum.energies))) / 2

    spectrum.add_moments(energy_resolution=tol)
    new_sampling_points = 2 * spectrum.num_moments

    assert old_sampling_points < new_sampling_points
    assert np.max(np.abs(np.diff(spectrum.energies))) < tol


# ### check _rescale


def test_rescale():
    ham = kwant.rmt.gaussian(dim)
    spectrum, filter_index = make_spectrum_and_peaks(ham, p)

    ham_operator, (a, b) = _rescale(ham, eps=0.05, v0=None, bounds=None)
    rescaled_eigvalues, rescaled_eigvectors = np.linalg.eigh((
        ham - b * np.identity(len(ham))) / a)

    assert np.all(-1 < rescaled_eigvalues) and np.all(rescaled_eigvalues < 1)

    """The test
    `assert np.max(np.abs(eigvectors - rescaled_eigvectors)) < TOL`
    is not good because eigenvectors could be defined up to a phase. It should
    compare that the eigenvectors are proportional to eachother. One way to
    test this is that the product gives a complex number in the unit circle."""
    eigvalues, eigvectors = np.linalg.eigh(ham)
    assert np.all(1 - np.abs(np.vdot(eigvectors, rescaled_eigvectors)) < TOL)
